import threading
import logging
l = logging.getLogger('pyvex.lifting.libvex')
l.setLevel(20) # Shut up

from .. import stmt
from . import Lifter, register, LiftingException
from .. import pvc, ffi
from ..enums import default_vex_archinfo, vex_endness_from_string

_libvex_lock = threading.Lock()

SUPPORTED = {'X86', 'AMD64', 'MIPS32', 'MIPS64', 'ARM', 'ARMEL', 'ARMHF', 'ARMCortexM', 'AARCH64', 'PPC32', 'PPC64', 'S390X'}

VEX_MAX_INSTRUCTIONS = 99
VEX_MAX_BYTES = 5000


class VexRegisterUpdates:
    VexRegUpd_INVALID = 0x700
    VexRegUpdSpAtMemAccess = 0x701
    VexRegUpdUnwindregsAtMemAccess = 0x702
    VexRegUpdAllregsAtMemAccess = 0x703
    VexRegUpdAllregsAtEachInsn = 0x704
    VexRegUpdLdAllregsAtEachInsn = 0x705


class LibVEXLifter(Lifter):

    __slots__ = ()

    REQUIRE_DATA_C = True

    @staticmethod
    def get_vex_log():
        return bytes(ffi.buffer(pvc.msg_buffer, pvc.msg_current_size)).decode() if pvc.msg_buffer != ffi.NULL else None

    def lift(self):
        if self.traceflags != 0 and l.getEffectiveLevel() > 20:
            l.setLevel(20)

        try:
            _libvex_lock.acquire()

            pvc.log_level = l.getEffectiveLevel()
            vex_arch = getattr(pvc, self.irsb.arch.vex_arch)

            if self.bytes_offset is None:
                self.bytes_offset = 0

            if self.max_bytes is None or self.max_bytes > VEX_MAX_BYTES:
                max_bytes = VEX_MAX_BYTES
            else:
                max_bytes = self.max_bytes

            if self.max_inst is None or self.max_inst > VEX_MAX_INSTRUCTIONS:
                max_inst = VEX_MAX_INSTRUCTIONS
            else:
                max_inst = self.max_inst

            strict_block_end = self.strict_block_end
            if strict_block_end is None:
                strict_block_end = True

            collect_data_refs = self.collect_data_refs
            if collect_data_refs is None:
                collect_data_refs = False

            if self.cross_insn_opt:
                px_control = VexRegisterUpdates.VexRegUpdUnwindregsAtMemAccess
            else:
                px_control = VexRegisterUpdates.VexRegUpdLdAllregsAtEachInsn

            self.irsb.arch.vex_archinfo['hwcache_info']['caches'] = ffi.NULL
            lift_r = pvc.vex_lift(vex_arch,
                                  self.irsb.arch.vex_archinfo,
                                  self.data + self.bytes_offset,
                                  self.irsb.addr,
                                  max_inst,
                                  max_bytes,
                                  self.opt_level,
                                  self.traceflags,
                                  self.allow_arch_optimizations,
                                  strict_block_end,
                                  collect_data_refs,
                                  px_control,
                                  self.bytes_offset,
                                  )
            log_str = self.get_vex_log()
            if lift_r == ffi.NULL:
                raise LiftingException("libvex: unknown error" if log_str is None else log_str)
            else:
                if log_str is not None:
                    l.info(log_str)

            self.irsb._from_c(lift_r, skip_stmts=self.skip_stmts)
            if self.irsb.size == 0:
                l.debug('raising lifting exception')
                raise LiftingException("libvex: could not decode any instructions @ 0x%x" % self.addr)
        finally:
            _libvex_lock.release()
            self.irsb.arch.vex_archinfo['hwcache_info']['caches'] = None


for arch_name in SUPPORTED:
    register(LibVEXLifter, arch_name)




import logging
import bitstring

from .vex_helper import *
from ..lifter import Lifter
from ...const import vex_int_class

l = logging.getLogger(__name__)

def is_empty(bitstrm):
    try:
        bitstrm.peek(1)
        return False
    except bitstring.ReadError:
        return True


class ParseError(Exception):
    pass


class GymratLifter(Lifter):
    """
    This is a base class for lifters that use Gymrat.
    For most architectures, all you need to do is subclass this, and set the property "instructions"
    to be a list of classes that define each instruction.
    By default, a lifter will decode instructions by attempting to instantiate every class until one works.
    This will use an IRSBCustomizer, which will, if it succeeds, add the appropriate VEX instructions to a pyvex IRSB.
    pyvex, when lifting a block of code for this architecture, will call the method "lift", which will produce the IRSB
    of the lifted code.
    """

    __slots__ = ('bitstrm', 'errors', 'thedata', )

    REQUIRE_DATA_PY = True
    instrs = None

    def __init__(self, *args):
        super().__init__(*args)
        self.bitstrm = None
        self.errors = None
        self.thedata = None

    def create_bitstrm(self):
        self.bitstrm = bitstring.ConstBitStream(bytes=self.thedata)

    def _decode_next_instruction(self, addr):
        # Try every instruction until one works
        for possible_instr in self.instrs:
            try:
                l.info("Trying " + possible_instr.name)
                return possible_instr(self.bitstrm, self.irsb.arch, addr)
            # a ParserError signals that this instruction did not match
            # we need to try other instructions, so we ignore this error
            except ParseError:
                pass #l.exception(repr(possible_instr))
            # if we are out of input, ignore.
            # there may be other, shorter instructions that still match,
            # so we continue with the loop
            except bitstring.ReadError:
                pass

        # If no instruction matches, log an error
        errorstr = 'Unknown instruction at bit position %d' % self.bitstrm.bitpos
        l.debug(errorstr)
        l.debug("Address: %#08x" % addr)

    def decode(self):
        try:
            self.create_bitstrm()
            count = 0
            disas = []
            addr = self.irsb.addr
            l.debug("Starting block at address: " + hex(addr))
            bytepos = self.bitstrm.bytepos


            while (not is_empty(self.bitstrm)):
                instr = self._decode_next_instruction(addr)
                if not instr: break
                disas.append(instr)
                l.debug("Matched " + instr.name)
                addr += self.bitstrm.bytepos - bytepos
                bytepos = self.bitstrm.bytepos
                count += 1
            return disas
        except Exception as e:
            self.errors = str(e)
            l.exception("Error decoding block at offset {:#x} (address {:#x}):".format(bytepos, addr))
            raise

    def lift(self, disassemble=False, dump_irsb=False):
        self.thedata = self.data[:self.max_bytes] if isinstance(self.data, (bytes, bytearray, memoryview)) else self.data[:self.max_bytes].encode()
        l.debug(repr(self.thedata))
        instructions = self.decode()

        if disassemble:
            return [instr.disassemble() for instr in instructions]
        self.irsb.jumpkind = JumpKind.Invalid
        irsb_c = IRSBCustomizer(self.irsb)
        l.debug("Decoding complete.")
        for i, instr in enumerate(instructions[:self.max_inst]):
            l.debug("Lifting instruction " + instr.name)
            instr(irsb_c, instructions[:i], instructions[i+1:])
            if irsb_c.irsb.jumpkind != JumpKind.Invalid:
                break
            elif (i+1) == self.max_inst: # if we are on our last iteration
                instr.jump(None, irsb_c.irsb.addr + irsb_c.irsb.size)
                break
        else:
            if len(irsb_c.irsb.statements) == 0:
                raise LiftingException('Could not decode any instructions')
            irsb_c.irsb.jumpkind = JumpKind.NoDecode
            dst = irsb_c.irsb.addr + irsb_c.irsb.size
            dst_ty = vex_int_class(irsb_c.irsb.arch.bits).type
            irsb_c.irsb.next = irsb_c.mkconst(dst, dst_ty)
        l.debug(self.irsb._pp_str())
        if dump_irsb:
            self.irsb.pp()
        return self.irsb

    def pp_disas(self):
        disasstr = ""
        insts = self.disassemble()
        for addr, name, args in insts:
            args_str = ",".join(str(a) for a in args)
            disasstr += "%0#08x:\t%s %s\n" % (addr, name, args_str)
        print(disasstr)

    def error(self):
        return self.errors

    def disassemble(self):
        return self.lift(disassemble=True)


from ...expr import *
from .. import LiftingException



"""
PyVEX provides an interface that translates binary code into the VEX intermediate represenation (IR).
For an introduction to VEX, take a look here: https://docs.angr.io/advanced-topics/ir
"""
from typing import NewType, Any

__version__ = (9, 0, 10281)

if bytes is str:
    raise Exception("This module is designed for python 3 only. Please install an older version to use python 2.")

import os
import sys
import hashlib
import pickle
import tempfile
import cffi
from .vex_ffi import ffi_str as _ffi_str
ffi = cffi.FFI()

import logging
logging.getLogger("pyvex").addHandler(logging.NullHandler())


def _locate_lib(module: str, library: str) -> str:
    """
    Attempt to find a native library without using pkg_resources, and only fall back to pkg_resources upon failures.
    This is because "import pkg_resources" is slow.

    :return:    The full path of the native library.
    """
    base_dir = os.path.dirname(__file__)
    attempt = os.path.join(base_dir, library)
    if os.path.isfile(attempt):
        return attempt

    import pkg_resources  # pylint:disable=import-outside-toplevel
    return pkg_resources.resource_filename(module, os.path.join('lib', library))


def _parse_ffi_str():
    hash = hashlib.md5(_ffi_str.encode("utf-8")).hexdigest()
    cache_location = os.path.join(tempfile.gettempdir(), f"pyvex_ffi_parser_cache.{hash}")

    if os.path.isfile(cache_location):
        # load the cache
        with open(cache_location, "rb") as f:
            cache = pickle.loads(f.read())
        ffi._parser._declarations = cache['_declarations']
        ffi._parser._int_constants = cache['_int_constants']
    else:
        ffi.cdef(_ffi_str)
        # cache the result
        cache = {
            '_declarations': ffi._parser._declarations,
            '_int_constants': ffi._parser._int_constants,
        }
        with open(cache_location, "wb") as f:
            f.write(pickle.dumps(cache))


def _find_c_lib():
    # Load the c library for calling into VEX
    if sys.platform in ('win32', 'cygwin'):
        library_file = 'pyvex.dll'
    elif sys.platform == 'darwin':
        library_file = "libpyvex.dylib"
    else:
        library_file = "libpyvex.so"

    pyvex_path = _locate_lib(__name__, os.path.join("lib", library_file))
    # parse _ffi_str and use cache if possible
    _parse_ffi_str()
    # RTLD_GLOBAL used for sim_unicorn.so
    lib = ffi.dlopen(pyvex_path)
    if not lib.vex_init():
        raise ImportError("libvex failed to initialize")
    # this looks up all the definitions (wtf)
    dir(lib)
    return lib

pvc = _find_c_lib() # type: Any # This should be properly typed, but this seems non trivial

# pylint: disable=wildcard-import
from .enums import *
from . import stmt, expr, const
from .block import IRSB, IRTypeEnv
from .expr import get_op_retty
from .const import tag_to_const_class, get_type_size, get_type_spec_size
from .lifting import lift, lifters
from .errors import PyVEXError

# aliases....
IRStmt = stmt
IRExpr = expr
IRConst = const



import logging
from ..util.lifter_helper import GymratLifter
from ..util.instr_helper import Instruction
from .. import register

l = logging.getLogger(__name__)

class Aarch64Instruction(Instruction): # pylint: disable=abstract-method
    # NOTE: WARNING: There is no MRS, MSR, SYSL in VEX's ARM implementation
    # You must use straight nasty hacks instead.
    pass

class Instruction_SYSL(Aarch64Instruction):
    name = "SYSL"
    bin_format = "1101010100101qqqnnnnmmmmppprrrrr"

    def compute_result(self): # pylint: disable=arguments-differ
        l.debug("Ignoring SYSL instruction at %#x.", self.addr)

class Instruction_MSR(Aarch64Instruction):
    name = "MSR"
    bin_format = "11010101000ioqqqnnnnmmmmppprrrrr"

    def compute_result(self): # pylint: disable=arguments-differ
        l.debug("Ignoring MSR instruction at %#x.", self.addr)

class Instruction_MRS(Aarch64Instruction):
    name = "MRS"
    bin_format = "110101010011opppnnnnmmmmppprrrrr"

    def compute_result(self): # pylint: disable=arguments-differ
        l.debug("Ignoring MRS instruction at %#x.", self.addr)

class AARCH64Spotter(GymratLifter):
    instrs = [
        Instruction_MRS,
        Instruction_MSR,
        Instruction_SYSL]

register(AARCH64Spotter, "ARM64")
register(AARCH64Spotter, "AARCH64")



import bitstring
import logging

from ..util.lifter_helper import GymratLifter
from ..util.instr_helper import Instruction, ParseError
from ..util import JumpKind, Type
from .. import register

l = logging.getLogger(__name__)


class ARMInstruction(Instruction): # pylint: disable=abstract-method

    # NOTE: WARNING: There is no CPSR in VEX's ARM implementation
    # You must use straight nasty hacks instead.

    # NOTE 2: Something is goofy w/r/t archinfo and VEX; cc_op3 is used in ccalls, but there's
    # no cc_op3 in archinfo, angr itself uses cc_depn instead.  We do the same.

    def match_instruction(self, data, bitstrm):
        """
        ARM Instructions are pretty dense, so let's do what we can to weed them out
        """
        if 'c' not in data or data['c'] == '1111':
            raise ParseError("Invalid ARM Instruction")

    def get_N(self):
        cc_op = self.get("cc_op", Type.int_32)
        cc_dep1 = self.get("cc_dep1", Type.int_32)
        cc_dep2 = self.get("cc_dep2", Type.int_32)
        cc_depn = self.get("cc_ndep", Type.int_32)
        return self.ccall(Type.int_32, "armg_calculate_flag_n", [cc_op, cc_dep1, cc_dep2, cc_depn])

    def get_C(self):
        cc_op = self.get("cc_op", Type.int_32)
        cc_dep1 = self.get("cc_dep1", Type.int_32)
        cc_dep2 = self.get("cc_dep2", Type.int_32)
        cc_depn = self.get("cc_ndep", Type.int_32)
        return self.ccall(Type.int_32, "armg_calculate_flag_c", [cc_op, cc_dep1, cc_dep2, cc_depn])

    def get_V(self):
        cc_op = self.get("cc_op", Type.int_32)
        cc_dep1 = self.get("cc_dep1", Type.int_32)
        cc_dep2 = self.get("cc_dep2", Type.int_32)
        cc_depn = self.get("cc_ndep", Type.int_32)
        return self.ccall(Type.int_32, "armg_calculate_flag_v", [cc_op, cc_dep1, cc_dep2, cc_depn])

    def get_Z(self):
        cc_op = self.get("cc_op", Type.int_32)
        cc_dep1 = self.get("cc_dep1", Type.int_32)
        cc_dep2 = self.get("cc_dep2", Type.int_32)
        cc_depn = self.get("cc_ndep", Type.int_32)
        return self.ccall(Type.int_32, "armg_calculate_flag_z", [cc_op.rdt, cc_dep1.rdt, cc_dep2.rdt, cc_depn.rdt])

    def evaluate_condition(self):
        # condition codes should be in 'c'
        cond = self.data['c']
        if cond == '0000':
            # equal, z set
            return self.get_Z() == 1
        elif cond == '0001':
            # not equal, Z clear
            return self.get_Z() == 0
        elif cond == '0010':
            # Carry, C set
            return self.get_C() == 1
        elif cond == '0011':
            # Carry Clear, C clear
            return self.get_C() == 0
        elif cond == '0100':
            # MI / neagative / N set
            return self.get_N() == 1
        elif cond == '0101':
            # PL / plus / positive / N clear
            return self.get_N() == 0
        elif cond == '0110':
            # VS / V set / Overflow
            return self.get_V() == 1
        elif cond == '0111':
            # VC / V Clear / no overflow
            return self.get_V() == 0
        elif cond == '1000':
            # Hi / unsigned higher / C set, Z clear
            return (self.get_C() == 1) & (self.get_Z() == 0)
        elif cond == '1001':
            # LS / C clear, Z set
            return (self.get_C() == 0) & (self.get_Z() == 1)
        elif cond == '1011':
            # LT / Less than / N != V
            return self.get_N() != self.get_V()
        elif cond == '1100':
            # GT / greater than / Z clear and (n == v)
            return (self.get_Z() == 1) & (self.get_N() != self.get_V())
        elif cond == '1101':
            # LE / less than or equal to / Z set OR (N != V)
            return (self.get_Z() == 1) | (self.get_N() != self.get_V())
        else:
            # No condition
            return None

    def _load_le_instr(self, bitstream: bitstring.ConstBitStream, numbits: int) -> str:
        # THUMB mode instructions swap endianness every two bytes!
        if (self.addr & 1) == 1 and numbits > 16:
            chunk = ""
            oldpos = bitstream.pos
            for _ in range(0, numbits, 16):
                chunk += bitstring.Bits(uint=bitstream.peek("uintle:%d" % 16), length=16).bin
                bitstream.pos += 16
            bitstream.pos = oldpos
            return chunk
        return super()._load_le_instr(bitstream, numbits)


class Instruction_MRC(ARMInstruction):
    name = "MRC"
    bin_format = 'cccc1110CCC1nnnnddddppppOOOOOOOO'
                 #11101110000100010001111100010000
    # c = cond
    # C = Coprocessor operation mode
    # d = CPd
    # O = Offset
    # p = CP#

    def compute_result(self): # pylint: disable=arguments-differ
        # TODO at least look at the conditionals
        # TODO Clobber the dst reg of MCR
        # TODO maybe treat coproc regs as simple storage (even though they are very much not)
        l.debug("Ignoring MRC instruction at %#x.", self.addr)


class Instruction_MCR(ARMInstruction):
    name = "MCR"
    bin_format = 'cccc1110CCC0nnnnddddppppOOOOOOOO'
                 #11101110000000010000111100010000
    # c = cond
    # C = Coprocessor operation mode
    # d = CPd
    # O = Offset
    # p = CP#

    def compute_result(self): # pylint: disable=arguments-differ
        # TODO at least look at the conditionals
        # TODO Clobber the dst reg of MCR
        # TODO maybe treat coproc regs as simple storage (even though they are very much not)
        l.debug("Ignoring MCR instruction at %#x.", self.addr)


class Instruction_MSR(ARMInstruction):
    name = "MSR"
    bin_format = 'cccc00i10d10xxxj1111ssssssssssss'
    #             11100011001000011111000010010001
    #             11100001011011111111000000000001

    def compute_result(self): # pylint: disable=arguments-differ
        l.debug("Ignoring MSR instruction at %#x. VEX cannot support this instruction. See pyvex/lifting/gym/arm_spotter.py", self.addr)


class Instruction_MRS(ARMInstruction):
    name = "MRS"
    bin_format = "cccc00010s001111dddd000000000000"

    def compute_result(self): # pylint: disable=arguments-differ
        l.debug("Ignoring MRS instruction at %#x. VEX cannot support this instruction. See pyvex/lifting/gym/arm_spotter.py", self.addr)


class Instruction_STM(ARMInstruction):
    name = "STM"
    bin_format = 'cccc100pu1w0bbbbrrrrrrrrrrrrrrrr'

    def match_instruction(self, data, bitstrm):
        # If we don't push anything, that's not real
        if int(data['r']) == 0:
            raise ParseError("Invalid STM instruction")
        return True

    def compute_result(self): # pylint: disable=arguments-differ
        l.warning("Ignoring STMxx ^ instruction at %#x. This mode is not implemented by VEX! See pyvex/lifting/gym/arm_spotter.py", self.addr)


class Instruction_LDM(ARMInstruction):
    name = "LDM"
    bin_format = 'cccc100PU1W1bbbbrrrrrrrrrrrrrrrr'

    def match_instruction(self, data, bitstrm):
        # If we don't push anything, that's not real
        if int(data['r']) == 0:
            raise ParseError("Invalid LDM instruction")
        return True

    def compute_result(self): # pylint: disable=arguments-differ
        # test if PC will be set. If so, the jumpkind of this block should be Ijk_Ret
        l.warning("Spotting an LDM instruction at %#x.  This is not fully tested.  Prepare for errors.", self.addr)
        #l.warning(repr(self.rawbits))
        #l.warning(repr(self.data))

        src_n = int(self.data['b'], 2)
        src = self.get(src_n, Type.int_32)

        for reg_num, bit in enumerate(self.data['r']):
            reg_num = 15 - reg_num
            if bit == '1':
                if self.data['P'] == '1':
                    if self.data['U'] == '0':
                        src += 4
                    else:
                        src -= 4
                val = self.load(src, Type.int_32)
                self.put(val, reg_num)
                if self.data['P'] == '0':
                    if self.data['U'] == '0':
                        src += 4
                    else:
                        src -= 4
                # If we touch PC, we're doing a RET!
                if reg_num == 15 and bit == '1':
                    cond = self.evaluate_condition()
                    if cond is not None:
                        self.jump(cond, val, JumpKind.Ret)
                    else:
                        self.jump(None, val, JumpKind.Ret)
        # Write-back
        if self.data['W'] == '1':
            self.put(src, src_n)


class Instruction_STC(ARMInstruction):
    name = 'STC'
    bin_format = 'cccc110PUNW0nnnnddddppppOOOOOOOO'

    def compute_result(self): # pylint: disable=arguments-differ
        # TODO At least look at the conditionals
        l.debug("Ignoring STC instruction at %#x.", self.addr)


class Instruction_STC_THUMB(ARMInstruction):
    name = 'STC'
    bin_format = '111c110PUNW0nnnnddddppppOOOOOOOO'

    def compute_result(self): # pylint: disable=arguments-differ
        # TODO At least look at the conditionals
        l.debug("Ignoring STC instruction at %#x.", self.addr)


class Instruction_LDC(ARMInstruction):
    name = 'LDC'
    bin_format = 'cccc110PUNW1nnnnddddppppOOOOOOOO'

    def compute_result(self): # pylint: disable=arguments-differ
        # TODO At least look at the conditionals
        # TODO Clobber the dest reg of LDC
        # TODO Maybe clobber the dst reg of CDP, if we're really adventurous
        l.debug("Ignoring LDC instruction at %#x.", self.addr)


class Instruction_LDC_THUMB(ARMInstruction):
    name = 'LDC'
    bin_format = '111c110PUNW1nnnnddddppppOOOOOOOO'

    def compute_result(self): # pylint: disable=arguments-differ
        # TODO At least look at the conditionals
        # TODO Clobber the dest reg of LDC
        # TODO Maybe clobber the dst reg of CDP, if we're really adventurous
        l.debug("Ignoring LDC instruction at %#x.", self.addr)


class Instruction_CDP(Instruction):
    name = "CDP"
    bin_format = 'cccc1110oooonnnnddddppppPPP0mmmm'
    # c = cond
    # d = CPd
    # O = Offset
    # p = CP#

    def compute_result(self): # pylint: disable=arguments-differ
        # TODO At least look at the conditionals
        # TODO Maybe clobber the dst reg of CDP, if we're really adventurous
        l.debug("Ignoring CDP instruction at %#x.", self.addr)


##
## Thumb! (ugh)
##

class ThumbInstruction(Instruction): # pylint: disable=abstract-method

    def mark_instruction_start(self):
        self.irsb_c.imark(self.addr-1, self.bytewidth, 1)


class Instruction_tCPSID(ThumbInstruction):
    name = 'CPSID'
    bin_format = '101101x0011x0010'

    def compute_result(self): # pylint: disable=arguments-differ
        # TODO haha lol yeah right
        l.debug("[thumb] Ignoring CPS instruction at %#x.", self.addr)

class Instruction_tMSR(ThumbInstruction):
    name = 'tMSR'
    bin_format = '10x0mmmmxxxxxxxx11110011100Rrrrr'

    def compute_result(self): # pylint: disable=arguments-differ
        dest_spec_reg = int(self.data['x'], 2)
        src_reg = int(self.data['r'], 2)

        # If 0, do not write the SPSR
        if self.data['R'] == '0':
            if dest_spec_reg == 8: #msp
                src = self.get(src_reg, Type.int_32)
                self.put(src, 'sp')
            elif dest_spec_reg == 16: #primask
                src = self.get(src_reg, Type.int_32)
                self.put(src, 'primask')
            else:
               l.warning("[thumb] tMSR at %#x is writing into an unsupported special register %#x. Ignoring the instruction. FixMe.", self.addr, dest_spec_reg)
        else:
            l.warning("[thumb] tMSR at %#x is writing SPSR. Ignoring the instruction. FixMe.", self.addr)
        l.warning("[thumb] Spotting an tMSR instruction at %#x.  This is not fully tested.  Prepare for errors." , self.addr)

class Instruction_tMRS(ThumbInstruction):
    name = 'tMRS'
    bin_format = '10x0mmmmxxxxxxxx11110011111Rrrrr'

    def compute_result(self): # pylint: disable=arguments-differ

        spec_reg = int(self.data['x'], 2)
        dest_reg = int(self.data['m'], 2)

        # Reading from CPSR
        if self.data['R'] == '0':
            # See special registers constants here:
            # https://github.com/aquynh/capstone/blob/45bec1a691e455b864f7e4d394711a467e5493dc/arch/ARM/ARMInstPrinter.c#L1654
            if spec_reg == 8:
                # We move the SP and call it a day.
                src = self.get("sp", Type.int_32)
                self.put(src, dest_reg)
            elif spec_reg == 16:
                src = self.get("primask", Type.int_32)
                self.put(src, dest_reg)
            else:
                l.warning("[thumb] tMRS at %#x is using the unsupported special register %#x. Ignoring the instruction. FixMe." , self.addr, spec_reg)
        else:
            l.warning("[thumb] tMRS at %#x is reading from SPSR. Ignoring the instruction. FixMe." , self.addr)
            l.debug("[thumb] Ignoring tMRS instruction at %#x.", self.addr)
        l.warning("[thumb] Spotting an tMRS instruction at %#x.  This is not fully tested.  Prepare for errors." , self.addr)


class Instruction_tDMB(ThumbInstruction):
    name = 'DMB'
    bin_format = '100011110101xxxx1111001110111111'
    def compute_result(self):  # pylint: disable=arguments-differ
        # TODO haha lol yeah right
        l.debug("[thumb] Ignoring DMB instruction at %#x.", self.addr)



class Instruction_WFI(ThumbInstruction):
    name = "WFI"
    bin_format = "10111111001a0000"
                 #1011111100110000

    def compute_result(self): # pylint: disable=arguments-differ
        l.debug("[thumb] Ignoring WFI instruction at %#x.", self.addr)


class ARMSpotter(GymratLifter):
    arm_instrs = [
        Instruction_MRC,
        Instruction_MCR,
        Instruction_MSR,
        Instruction_MRS,
        Instruction_STM,
        Instruction_LDM,
        Instruction_STC,
        Instruction_LDC,
        Instruction_CDP,
    ]
    thumb_instrs = [Instruction_tCPSID,
                    Instruction_tMSR,
                    Instruction_tMRS,
                    Instruction_WFI,
                    Instruction_tDMB,
                    Instruction_STC_THUMB,
                    Instruction_LDC_THUMB,
                    ]

    def __init__(self, *args):
        super().__init__(*args)
        self.thumb: bool = False

    def lift(self, disassemble=False, dump_irsb=False):
        if self.irsb.addr & 1:
            # Thumb!
            self.instrs = self.thumb_instrs
            self.thumb = True
        else:
            self.instrs = self.arm_instrs
            self.thumb = False
        super().lift(disassemble, dump_irsb)

register(ARMSpotter, "ARM")
register(ARMSpotter, "ARMEL")
register(ARMSpotter, "ARMHF")
register(ARMSpotter, "ARMCortexM")



import re
import os
import sys
import cffi
import subprocess
import platform

import logging
l = logging.getLogger('cffier')
l.setLevel(logging.DEBUG)


def find_good_scan(questionable):
    known_good = []

    end_line = len(questionable)

    while len(questionable):
        ffi = cffi.FFI()
        l.debug("scan - trying %d good and %d questionable", len(known_good), len(questionable))

        candidate = known_good + questionable[:end_line]
        failed_line = -1

        try:
            ffi.cdef('\n'.join(candidate))

            known_good = candidate
            questionable = questionable[end_line:]
            end_line = len(questionable)
        except AssertionError:
            questionable = questionable[1:]
            end_line = len(questionable)
        except cffi.CDefError as e:
            if '<cdef source string>' in str(e):
                failed_line = int(str(e).split('\n')[-1].split(':')[1])-1
            elif str(e).count(':') >= 2:
                failed_line = int(str(e).split('\n')[1].split(':')[1])
                failed_line_description = str(e).split('\n')[0]
                idx1 = failed_line_description.index('"')
                idx2 = failed_line_description.rindex('"')
                failed_reason = failed_line_description[idx1+1:idx2]

                for i in range(failed_line, -1, -1):
                    if failed_reason in candidate[i]:
                        failed_line = i
            elif 'unrecognized construct' in str(e):
                failed_line = int(str(e).split()[1][:-1])-1
            elif 'end of input' in str(e):
                end_line -= 1
            else:
                raise Exception("Unknown error")
        except cffi.FFIError as e:
            if str(e).count(':') >= 2:
                failed_line = int(str(e).split('\n')[0].split(':')[1])-1
            else:
                raise Exception("Unknown error")

        if failed_line != -1:
            end_line = failed_line-len(known_good)

        if end_line == 0:
            questionable = questionable[1:]
            end_line = len(questionable)
    return known_good


def doit(vex_path):
    cpplist = ['cl', 'cpp']
    cpp = os.getenv("CPP")
    if cpp:
        cpplist.insert(0, cpp)
    if platform.system() == 'Darwin':
        cpplist.insert(0, "clang")

    errs = []
    for cpp in cpplist:
        cmd = [cpp, '-I' + vex_path, os.path.join("pyvex_c", "pyvex.h")]
        if cpp in ('cl', 'clang', 'gcc', 'cc', 'clang++', 'g++'):
            cmd.append("-E")
        try:
            p = subprocess.Popen(cmd,
                                 stdout=subprocess.PIPE,
                                 stderr=subprocess.PIPE)
            header, stderr = p.communicate()
            try:
                header = header.decode("utf-8")
                stderr = stderr.decode("utf-8")
            except UnicodeDecodeError:
                # They don't have to be unicode on Windows
                pass

            if not header.strip() or p.returncode != 0:
                errs.append((" ".join(cmd), p.returncode, stderr))
                continue
            else:
                break
        except OSError:
            errs.append((" ".join(cmd), -1, "does not exist"))
            continue
    else:
        l.warning("failed commands:\n" +
                  "\n".join("{} ({}) -- {}".format(*e) for e in errs))
        raise Exception(
            "Couldn't process pyvex headers." +
            "Please set CPP environmental variable to local path of \"cpp\"." +
            "Note that \"cpp\" and \"g++\" are different."
        )
    # header = vex_pp + pyvex_pp

    linesep = '\r\n' if '\r\n' in header else '\n'
    ffi_text = linesep.join(line for line in header.split(linesep) if '#' not in line and line.strip() != '' and 'jmp_buf' not in line and not ('=' in line and ';' in line))
    ffi_text = re.sub('\{\s*\} NoOp;', '{ int DONOTUSE; } NoOp;', ffi_text)
    ffi_text = re.sub('__attribute__\s*\(.*\)', '', ffi_text)
    ffi_text = re.sub('__declspec\s*\([^\)]*\)', '', ffi_text)
    ffi_text = ffi_text.replace('__const', 'const')
    ffi_text = ffi_text.replace('__inline', '')
    ffi_text = ffi_text.replace('__w64', '')
    ffi_text = ffi_text.replace('__cdecl', '')
    ffi_text = ffi_text.replace('__int64', 'long')
    ffi_lines = ffi_text.split(linesep)

    good = find_good_scan(ffi_lines)
    good += ['extern VexControl vex_control;']

    with open('pyvex/vex_ffi.py', 'w') as fp:
        fp.write('ffi_str = """' + '\n'.join(good) + '"""\n')
        fp.write('guest_offsets = ' + repr(get_guest_offsets(vex_path)) + '\n')

def get_guest_offsets(vex_path):
    fname = os.path.join(vex_path, 'libvex_guest_offsets.h')
    out = {}
    with open(fname) as fp:
        for line in fp:
            if line.startswith('#define'):
                _, names, val = line.split()
                val = int(val, 0)
                assert names.startswith('OFFSET_')
                _, arch, reg = names.split('_', 2)
                out[(arch, reg.lower())] = val
    return out

if __name__ == '__main__':
    import sys
    logging.basicConfig(level=logging.DEBUG)
    doit(sys.argv[1])
